TEST_VC_DOCUMENT_NESTED_REVEAL = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "issuer": {},
    "issuanceDate": {},
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "@explicit": True,
    "credentialSubject": {"@explicit": True, "degree": {"@explicit": True, "name": {}}},
}
